@@ -1,11 +1,20 @@
 import copy
 import datetime
-import typing
 import uuid
+from typing import List, Optional, Union
+
+from marshmallow import Schema
+from marshmallow import fields as schema_fields
 
 from commercetools import schemas, types
+from commercetools.testing import utils
 from commercetools.testing.abstract import BaseModel, ServiceBackend
-from commercetools.testing.utils import custom_fields_from_draft
+from commercetools.testing.utils import (
+    create_commercetools_response,
+    custom_fields_from_draft,
+    get_product_variants,
+    parse_request_params,
+)
 
 
 class ProductsModel(BaseModel):
@@ -14,7 +23,7 @@ class ProductsModel(BaseModel):
     _unique_values = ["key"]
 
     def _create_from_draft(
-        self, draft: types.ProductDraft, id: typing.Optional[str] = None
+        self, draft: types.ProductDraft, id: Optional[str] = None
     ) -> types.Product:
         object_id = str(uuid.UUID(id) if id is not None else uuid.uuid4())
 
@@ -55,11 +64,11 @@ class ProductsModel(BaseModel):
         self, draft: types.ProductVariantDraft
     ) -> types.ProductVariant:
 
-        assets: typing.Optional[typing.List[types.Asset]] = None
+        assets: Optional[List[types.Asset]] = None
         if draft.assets:
             assets = self._create_assets_from_draft(draft.assets)
 
-        prices: typing.Optional[typing.List[types.Price]] = None
+        prices: Optional[List[types.Price]] = None
         if draft.prices:
             prices = self._create_prices_from_draft(draft.prices)
 
@@ -79,11 +88,11 @@ class ProductsModel(BaseModel):
         )
 
     def _create_assets_from_draft(
-        self, drafts: typing.List[types.AssetDraft]
-    ) -> typing.List[types.Asset]:
-        assets: typing.List[types.Asset] = []
+        self, drafts: List[types.AssetDraft]
+    ) -> List[types.Asset]:
+        assets: List[types.Asset] = []
         for draft in drafts:
-            custom: typing.Optional[types.CustomFields] = None
+            custom: Optional[types.CustomFields] = None
             if draft.custom:
                 custom = custom_fields_from_draft(self._storage, draft.custom)
 
@@ -99,9 +108,9 @@ class ProductsModel(BaseModel):
         return assets
 
     def _create_prices_from_draft(
-        self, drafts: typing.List[types.PriceDraft]
-    ) -> typing.List[types.Price]:
-        prices: typing.List[types.Price] = []
+        self, drafts: List[types.PriceDraft]
+    ) -> List[types.Price]:
+        prices: List[types.Price] = []
         for draft in drafts:
             custom = None
             if draft.custom:
@@ -122,8 +131,8 @@ class ProductsModel(BaseModel):
         return prices
 
     def _create_price_from_draft(
-        self, draft: typing.Optional[types.TypedMoneyDraft]
-    ) -> typing.Optional[types.TypedMoney]:
+        self, draft: Union[types.Money, Optional[types.TypedMoneyDraft]]
+    ) -> Optional[types.TypedMoney]:
         if draft is None:
             return None
 
@@ -149,17 +158,33 @@ class ProductsModel(BaseModel):
             )
 
 
-def _get_target_obj(obj: dict, action: types.ProductUpdateAction):
-    staged = getattr(action, "staged", False)
+def _get_target_obj(obj: dict, staged: bool):
     if not staged and obj["masterData"]["current"]:
         return obj["masterData"]["current"]
     return obj["masterData"]["staged"]
 
 
+def _get_variant(data_object: dict, *, id: int = "", sku: str = "") -> Optional[dict]:
+    if not data_object:
+        return None
+
+    variants = [data_object["masterVariant"]]
+    if data_object["variants"]:
+        variants += data_object["variants"]
+
+    for variant in variants:
+        if id and id == variant["id"]:
+            return variant
+        if sku and sku == variant["sku"]:
+            return variant
+
+    return None
+
+
 def _update_productdata_attr(dst: str, src: str):
     def updater(self, obj: dict, action: types.ProductUpdateAction):
         value = getattr(action, src)
-        target_obj = _get_target_obj(obj, action)
+        target_obj = _get_target_obj(obj, getattr(action, "staged", False))
 
         if target_obj[dst] != value:
             new = copy.deepcopy(obj)
@@ -175,16 +200,9 @@ def _set_attribute_action():
         staged = getattr(action, "staged", False)
         new = copy.deepcopy(obj)
 
-        target_obj = _get_target_obj(new, action)
-
-        variants = [target_obj["masterVariant"]]
-        if target_obj["variants"]:
-            variants += target_obj["variants"]
-
-        for variant in variants:
-            if action.variant_id != variant["id"]:
-                continue
-
+        target_obj = _get_target_obj(new, staged)
+        variant = _get_variant(target_obj, id=action.variant_id)
+        if variant:
             for attr in variant["attributes"]:
                 if attr["name"] == action.name:
                     attr["value"] = action.value
@@ -195,7 +213,6 @@ def _set_attribute_action():
             variant["attributes"].append(
                 attr_schema.dump(types.Attribute(name=action.name, value=action.value))
             )
-            break
 
         return new
 
@@ -217,7 +234,7 @@ def _add_variant_action():
         schema = schemas.ProductVariantSchema()
 
         new = copy.deepcopy(obj)
-        target_obj = _get_target_obj(new, action)
+        target_obj = _get_target_obj(new, staged=getattr(action, "staged", True))
         if not target_obj["variants"]:
             target_obj["variants"] = []
         target_obj["variants"].append(schema.dump(variant))
@@ -227,6 +244,120 @@ def _add_variant_action():
     return updater
 
 
+def _publish_product_action():
+    def updater(self, obj: dict, action: types.ProductPublishAction):
+        new = copy.deepcopy(obj)
+        # not implemented scopes right now.
+        if new["masterData"].get("staged"):
+            new["masterData"]["current"] = new["masterData"]["staged"]
+            del new["masterData"]["staged"]
+        new["masterData"]["hasStagedChanges"] = False
+        new["masterData"]["published"] = True
+        return new
+
+    return updater
+
+
+def convert_draft_price(
+    price_draft: types.PriceDraft, price_id: str = None
+) -> types.Price:
+    tiers: Optional[List[types.PriceTier]] = None
+    if price_draft.tiers:
+        tiers = [utils.create_from_draft(tier) for tier in price_draft.tiers]
+    return types.Price(
+        id=price_id or str(uuid.uuid4()),
+        country=price_draft.country,
+        channel=price_draft.channel,
+        value=utils._money_to_typed(price_draft.value),
+        valid_from=price_draft.valid_from,
+        valid_until=price_draft.valid_until,
+        discounted=price_draft.discounted,
+        custom=utils.create_from_draft(price_draft.custom),
+        tiers=tiers,
+    )
+
+
+def _set_product_prices():
+    def updater(self, obj: dict, action: types.ProductSetPricesAction):
+        new = copy.deepcopy(obj)
+        target_obj = _get_target_obj(new, getattr(action, "staged", True))
+        prices = []
+        for price_draft in action.prices:
+            price = convert_draft_price(price_draft)
+            prices.append(price)
+
+        schema = schemas.PriceSchema()
+        for variant in get_product_variants(target_obj):
+            if variant["sku"] == action.sku:
+                variant["prices"] = schema.dump(prices, many=True)
+        if action.staged:
+            new["masterData"]["hasStagedChanges"] = True
+        return new
+
+    return updater
+
+
+def _change_price():
+    def updater(self, obj: dict, action: types.ProductChangePriceAction):
+        new = copy.deepcopy(obj)
+        staged = action.staged
+        if staged is None:
+            staged = True
+        target_obj = _get_target_obj(new, staged)
+        changed_price = convert_draft_price(action.price, action.price_id)
+        schema = schemas.PriceSchema()
+
+        found_price = True
+        for variant in get_product_variants(target_obj):
+            for index, price in enumerate(variant["prices"]):
+                if price["id"] == action.price_id:
+                    variant["prices"][index] = schema.dump(changed_price)
+                    found_price = True
+                    break
+        if not found_price:
+            raise ValueError("Could not find price with id %s" % action.price_id)
+        if staged:
+            new["masterData"]["hasStagedChanges"] = True
+        return new
+
+    return updater
+
+
+def _add_price():
+    def updater(self, obj: dict, action: types.ProductAddPriceAction):
+        new = copy.deepcopy(obj)
+        staged = action.staged
+        if staged is None:
+            staged = True
+        target_obj = _get_target_obj(new, staged)
+        new_price = convert_draft_price(action.price)
+        schema = schemas.PriceSchema()
+
+        found_sku = False
+        for variant in get_product_variants(target_obj):
+            if variant["sku"] == action.sku:
+                if "prices" not in variant:
+                    variant["prices"] = []
+                elif not variant["prices"]:
+                    variant["prices"] = []
+                variant["prices"].append(schema.dump(new_price))
+                found_sku = True
+                break
+        if not found_sku:
+            raise ValueError("Could not find sku %s" % action.sku)
+        if staged:
+            new["masterData"]["hasStagedChanges"] = True
+        return new
+
+    return updater
+
+
+class UploadImageQuerySchema(Schema):
+    staged = schema_fields.Bool()
+    filename = schema_fields.Field()
+    sku = schema_fields.Field()
+
+
 class ProductsBackend(ServiceBackend):
     service_path = "products"
     model_class = ProductsModel
@@ -244,6 +375,7 @@ class ProductsBackend(ServiceBackend):
             ("^(?P<id>[^/]+)$", "POST", self.update_by_id),
             ("^(?P<id>[^/]+)$", "DELETE", self.delete_by_id),
             ("^key=(?P<key>[^/]+)$", "DELETE", self.delete_by_key),
+            ("^(?P<id>[^/]+)\/images$", "POST", self.upload_image),
         ]
 
     # Fixme: use decorator for this
@@ -251,4 +383,31 @@ class ProductsBackend(ServiceBackend):
         "changeSlug": _update_productdata_attr("slug", "slug"),
         "setAttribute": _set_attribute_action(),
         "addVariant": _add_variant_action(),
+        "setPrices": _set_product_prices(),
+        "changePrice": _change_price(),
+        "addPrice": _add_price(),
+        "publish": _publish_product_action(),
     }
+
+    def upload_image(self, request, id):
+        obj = self.model.get_by_id(id)
+        if not obj:
+            return create_commercetools_response(request, status_code=404)
+
+        obj = copy.deepcopy(obj)
+
+        params = parse_request_params(UploadImageQuerySchema, request)
+        target = _get_target_obj(obj, staged=params["staged"])
+        filename = params["filename"]
+
+        variant = _get_variant(target, sku=params["sku"])
+        if not variant["images"]:
+            variant["images"] = []
+        image_schema = schemas.ImageSchema()
+        variant["images"].append(
+            image_schema.dump(
+                types.Image(url=f"cdn.commercetools.local/detail-{filename}")
+            )
+        )
+        self.model.save(obj)
+        return create_commercetools_response(request, json=obj)
