@@ -1,3 +1,4 @@
+import copy
 import datetime
 import typing
 import uuid
@@ -22,9 +23,20 @@ class StoresModel(BaseModel):
             key=draft.key,
             name=draft.name,
             created_at=datetime.datetime.now(datetime.timezone.utc),
+            languages=draft.languages,
         )
 
 
+def set_languages():
+    def updater(self, obj, action):
+        value = getattr(action, "languages")
+        new = copy.deepcopy(obj)
+        new["languages"] = value
+        return new
+
+    return updater
+
+
 class StoresBackend(ServiceBackend):
     service_path = "stores"
     model_class = StoresModel
@@ -45,4 +57,7 @@ class StoresBackend(ServiceBackend):
             ("^key=(?P<key>[^/]+)$", "POST", self.update_by_key),
         ]
 
-    _actions = {"setName": update_attribute("name", "name")}
+    _actions = {
+        "setName": update_attribute("name", "name"),
+        "setLanguages": set_languages(),
+    }
